#include <sys/types.h>
#include <sys/stat.h>

#include <err.h>
#include <fcntl.h>
#include <limits.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include "arg.h"
#include "block.h"
#include "config.h"
#include "key.h"
#include "lock.h"
#include "misc.h"
#include "snap.h"
#include "state.h"

struct param param;
int verbose;
char *argv0;

static void
loadstate(char *repo)
{
	char path[PATH_MAX];
	int fd;

	if (snprintf(path, sizeof(path), "%s/state", repo) >= sizeof(path))
		errx(1, "snprintf: %s: path too long", path);
	fd = open(path, O_RDONLY);
	if (fd < 0)
		err(1, "open: %s", path);
	if (readstate(fd, &param) < 0)
		printerr("readstate: %s", path);
	if (close(fd) < 0)
		err(1, "close: %s", path);
}

static void
loadkey(char *keyfile)
{
	int fd;

	if (keyfile == NULL)
		return;

	fd = open(keyfile, O_RDONLY);
	if (fd < 0)
		err(1, "open: %s", keyfile);
	if (readkey(fd, param.key, sizeof(param.key)) < 0)
		printerr("readkey: %s", keyfile);
	param.keyloaded = 1;
	if (close(fd) < 0)
		err(1, "close: %s", keyfile);
}

static void
unpack(struct sctx *sctx, struct bctx *bctx)
{
	unsigned char md[MDSIZE];
	void *buf;
	int sn;

	buf = malloc(BSIZEMAX);
	if (buf == NULL)
		err(1, "malloc");
	while ((sn = sget(sctx, md)) == MDSIZE) {
		size_t bn = BSIZEMAX;

		if (bget(bctx, md, buf, &bn) < 0)
			printerr("bget");
		if (xwrite(1, buf, bn) != bn)
			err(1, "xwrite");
	}
	if (sn < 0)
		printerr("sget");
	free(buf);
}

static void
usage(void)
{
	fprintf(stderr, "usage: %s [-v] [-k keyfile] [-r repo] name\n", argv0);
	exit(1);
}

int
main(int argc, char *argv[])
{
	char spath[PATH_MAX];
	char bpath[PATH_MAX];
	struct sctx *sctx;
	struct bctx *bctx;
	char *keyfile = NULL;
	char *repo = ".";
	int lfd;

	ARGBEGIN {
	case 'k':
		keyfile = EARGF(usage());
		break;
	case 'r':
		repo = EARGF(usage());
		break;
	case 'v':
		verbose++;
		break;
	default:
		usage();
	} ARGEND

	if (argc != 1)
		usage();

	if (snprintf(spath, sizeof(spath), "%s/archive/%s",
	             repo, argv[0]) >= sizeof(spath))
		errx(1, "snprintf: %s: path too long", spath);
	if (snprintf(bpath, sizeof(bpath), "%s/storage",
	             repo) >= sizeof(bpath))
		errx(1, "snprintf: %s: path too long", bpath);

	if ((lfd = lockrepo(repo)) < 0)
		errx(1, "failed to lock repository");

	loadkey(keyfile);
	loadstate(repo);

	if (sopen(spath, S_READ, 0600, &sctx) < 0)
		printerr("sopen: %s", spath);
	if (bopen(bpath, B_READ, 0600, &bctx) < 0)
		printerr("bopen: %s", bpath);

	unpack(sctx, bctx);

	if (bclose(bctx) < 0)
		printerr("bclose: %s", bpath);
	if (sclose(sctx) < 0)
		printerr("sclose: %s", spath);

	if (unlockrepo(lfd) < 0)
		errx(1, "failed to unlock repository");
	
	return 0;
}
